# -*- coding: utf-8 -*-

if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())

from numba import jit, njit
import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.interpolate import interp1d
from frm.pricing_engine.monte_carlo_generic import generate_rand_nbs
from frm.pricing_engine.monte_carlo_generic import MAX_SIMULATIONS_PER_LOOP

import time
from prettytable import PrettyTable


def clewlow_strickland_1_factor(forward_curve, nb_simulations, segments_per_day, T, alpha, sigma):
    """
    Simulates spot prices using the Clewlow-Strickland one-factor model based on a provided forward curve.

    Parameters
    ----------
    forward_curve : numpy.ndarray
        A two-dimensional array where the first column contains time steps and the second column contains the corresponding forward prices.
    nb_simulations : int
        Number of simulation paths to generate.
    segments_per_day : int
        Number of extra simulation steps per forward curve step; higher values increase the model's temporal resolution.
    T : float
        Time horizon in days for the simulation.
    alpha : float
        Mean reversion speed of the model.
    sigma : float
        Volatility of the underlying asset.

    Returns
    -------
    spot_px_result : numpy.ndarray
        A two-dimensional array of simulated spot prices. Each row corresponds to a time step, and each column to a simulation path.
    
    Reference:
    Ahmos Sansom (2024). Clewlow and Strickland Commodity one factor spot model 
    (https://www.mathworks.com/matlabcentral/fileexchange/26969-clewlow-and-strickland-commodity-one-factor-spot-model), 
    MATLAB Central File Exchange. Retrieved April 28, 2024.        
    
    ##### To do #####
    Currently the code assumes the forward curve step has a constant increment.
    Want it to support a generic structure where the increment may change and may be denomintated in days, months or years. 
    The code simply discretises it generally. Need to figure out if it should discretise evenly from t=0 to T=T or discretize based on the forward curve granularity (which may change)
    
    Probably easiest to simulate the entire forward curve, and not include the T input that Atmos did. 
    The user can process the simulation results after.
    
    Likely the easiest way is to add extra granuarity between the timesteps provided by the user. 
    That gives the user more power as well - they can apply interpolation if they want rather than us assuming they want it. 
    
    May not be relevant for this function but days is likely the best data structure for functions to expect your data. 
    Easiest to convert to anything else given a year frac function 1st calls a day count function. 
    """

    nb_simulations = int(nb_simulations)
    segments_per_day = int(segments_per_day)
    
    a = alpha
    σ = sigma

    # this would need to be adjusted to the time between t and t+1 based on the forward curve
    dt = 1.0 / segments_per_day 
    
    
    
    nb_steps = T * segments_per_day # The Clewlow Strickland 1 Factor model requires granular discretisation in order to converge
    tau = (np.arange(nb_steps+1) * dt)
    
    f = interp1d(forward_curve[:,0], forward_curve[:,1], kind='linear') # Interpolate the forward curve based on the steps 
    forward_curve_interp = f(tau)
    index_spot = np.arange(T+1) * segments_per_day
    ln_forward_curve_interp = np.log(forward_curve_interp)
    
    nb_simulations_per_loop = int(min(nb_simulations, MAX_SIMULATIONS_PER_LOOP // nb_steps))
    nb_loops = (nb_simulations + nb_simulations_per_loop - 1) // nb_simulations_per_loop
    
    spot_px_result = np.zeros((T+1, nb_simulations))
    ln_spot_px = np.zeros((nb_steps+1, nb_simulations_per_loop))
    ln_spot_px[0, :] = ln_forward_curve_interp[0] # Set t=0 to the current spot price
    
    # These terms evaluate to constants hence can be calculated outside the monte carlo loop
    term1 = (ln_forward_curve_interp[1:] - ln_forward_curve_interp[:-1]) / dt
    term1 = np.concatenate(([np.nan], term1)) # Prefix with nan for t=0
    term3 = -0.25 * σ**2 * (1.0 + np.exp(-2.0 * a * tau[:-1]))
    term3 = np.concatenate(([np.nan], term3)) # Prefix with nan for t=0    

    for j in range(nb_loops):        
        rand_nbs = generate_rand_nbs(nb_steps=nb_steps, nb_rand_vars=1, nb_simulations=nb_simulations_per_loop, flag_apply_antithetic_variates=False, random_seed=j)

        for i in range(1, ln_spot_px.shape[0]):
            term2 = a * (ln_forward_curve_interp[i-1] - ln_spot_px[i-1,:])
            ln_spot_px[i,:] = ln_spot_px[i-1,:] + (term1[i] + term2 + term3[i]) * dt + rand_nbs[i-1,0,:] * σ * np.sqrt(dt)   

        spot_px = np.exp(ln_spot_px)        
        
        idx_start = j*nb_simulations_per_loop
        idx_end = min((j+1)*nb_simulations_per_loop, nb_simulations)
        idx = np.arange(idx_start,idx_end)
        
        if j < (nb_loops-1):    
            spot_px_result[:, idx] = spot_px[index_spot, :]
        else:
            index_spot_2d = index_spot[:, np.newaxis]
            mask = np.arange(0,spot_px.shape[1]) < (idx_end-idx_start)
            spot_px_result[:, idx] = spot_px[index_spot_2d, mask]

    return spot_px_result
    

if __name__ == "__main__":
    # This script reproduces the results in "ClewlowStricklandOneFactorSDE.pdf"on pages 11-12 sourced from:
    # Ahmos Sansom (2024). Clewlow and Strickland Commodity one factor spot model 
    # (https://www.mathworks.com/matlabcentral/fileexchange/26969-clewlow-and-strickland-commodity-one-factor-spot-model), 
    # MATLAB Central File Exchange. Retrieved April 28, 2024.        
        
    np.random.seed(0)
    forward_curve_arr = np.array([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 256, 257, 258, 259, 260, 261, 262, 263, 264, 265, 266, 267, 268, 269, 270, 271, 272, 273, 274, 275, 276, 277, 278, 279, 280, 281, 282, 283, 284, 285, 286, 287, 288, 289, 290, 291, 292, 293, 294, 295, 296, 297, 298, 299, 300, 301, 302, 303, 304, 305, 306, 307, 308, 309, 310, 311, 312, 313, 314, 315, 316, 317, 318, 319, 320, 321, 322, 323, 324, 325, 326, 327, 328, 329, 330, 331, 332, 333, 334, 335, 336, 337, 338, 339, 340, 341, 342, 343, 344, 345, 346, 347, 348, 349, 350, 351, 352, 353, 354, 355, 356, 357, 358, 359, 360, 361, 362, 363, 364, 365, 366, 367, 368, 369, 370, 371, 372, 373, 374, 375, 376, 377, 378, 379, 380, 381, 382, 383, 384, 385, 386, 387, 388, 389, 390, 391, 392, 393, 394, 395, 396, 397, 398, 399, 400, 401, 402, 403, 404, 405, 406, 407, 408, 409, 410, 411, 412, 413, 414, 415, 416, 417, 418, 419, 420, 421, 422, 423, 424, 425, 426, 427, 428, 429, 430, 431, 432, 433, 434, 435, 436, 437, 438, 439, 440, 441, 442, 443, 444, 445, 446, 447, 448, 449, 450, 451, 452, 453, 454, 455, 456, 457, 458, 459, 460, 461, 462, 463, 464, 465, 466, 467, 468, 469, 470, 471, 472, 473, 474, 475, 476, 477, 478, 479, 480, 481, 482, 483, 484, 485, 486, 487, 488, 489, 490, 491, 492, 493, 494, 495, 496, 497, 498, 499, 500, 501, 502, 503, 504, 505, 506, 507, 508, 509, 510, 511, 512, 513, 514, 515, 516, 517, 518, 519, 520, 521, 522, 523, 524, 525, 526, 527, 528, 529, 530, 531, 532, 533, 534, 535, 536, 537, 538, 539, 540, 541, 542, 543, 544, 545, 546, 547, 548, 549, 550, 551, 552, 553, 554, 555, 556, 557, 558, 559, 560, 561, 562, 563, 564, 565, 566, 567, 568, 569, 570, 571, 572, 573, 574, 575, 576, 577, 578, 579, 580, 581, 582, 583, 584, 585, 586, 587, 588, 589, 590, 591, 592, 593, 594, 595, 596, 597, 598, 599, 600, 601, 602, 603, 604, 605, 606, 607, 608, 609, 610, 611, 612, 613, 614, 615, 616, 617, 618, 619, 620, 621, 622, 623, 624, 625, 626, 627, 628, 629, 630, 631, 632, 633, 634, 635, 636, 637, 638, 639, 640, 641, 642, 643, 644, 645, 646, 647, 648, 649, 650, 651, 652, 653, 654, 655, 656, 657, 658, 659, 660, 661, 662, 663, 664, 665, 666, 667, 668, 669, 670, 671, 672, 673, 674, 675, 676, 677, 678, 679, 680, 681, 682, 683, 684, 685, 686, 687, 688, 689, 690, 691, 692, 693, 694, 695, 696, 697, 698, 699, 700, 701, 702, 703, 704, 705, 706, 707, 708, 709, 710, 711, 712, 713, 714, 715, 716, 717, 718, 719, 720, 721, 722, 723, 724, 725, 726, 727, 728, 729, 730, 731, 732, 733, 734, 735, 736, 737, 738, 739, 740, 741, 742, 743, 744, 745, 746, 747, 748, 749, 750, 751, 752, 753, 754, 755, 756, 757, 758, 759, 760],
                                  [53.50763333, 53.50763333, 50.49227708, 57.24444375, 61.88971875, 61.88971875, 61.88971875, 58.03241042, 55.75607083, 52.79759792, 58.98938125, 59.91974583, 59.91974583, 59.91974583, 57.52826042, 53.85879792, 50.833875, 59.50622708, 60.4127375, 60.4127375, 60.4127375, 58.06151667, 52.52509167, 49.56778125, 59.31270417, 60.20943333, 60.20943333, 60.20943333, 57.96403958, 53.37019375, 50.35380833, 57.58897917, 58.45470833, 58.45470833, 58.45470833, 56.35202292, 53.10617708, 50.08982917, 57.86863125, 58.78307083, 58.78307083, 58.78307083, 56.18114792, 58.99247292, 55.60075833, 56.67780208, 57.55663125, 57.55663125, 57.55663125, 55.1760625, 57.29396042, 54.00494792, 55.572525, 56.41594375, 56.41594375, 56.41594375, 54.26176042, 55.39162083, 52.21848958, 54.55282292, 55.36101458, 55.36101458, 55.36101458, 53.43822708, 53.28542292, 50.24140417, 53.61865, 54.39183542, 54.39183542, 54.39183542, 52.70545833, 50.95862917, 48.36986522, 52.75213125, 53.49079375, 53.49079375, 53.49079375, 52.04181458, 48.65551042, 45.89637083, 51.62509375, 52.32942083, 52.32942083, 52.32942083, 46.74299792, 46.73906875, 44.09534167, 44.09903125, 51.06860833, 51.06860833, 51.06860833, 49.95054167, 45.24201875, 42.68543333, 49.15257917, 49.81375, 49.81375, 49.81375, 48.76670417, 44.16434167, 41.66664375, 47.58353958, 48.22882292, 48.22882292, 48.22882292, 47.19777917, 43.41317917, 40.95285417, 40.81748958, 46.8231, 46.8231, 46.8231, 45.78414583, 42.79063542, 40.36055, 45.61413958, 46.23938958, 46.23938958, 46.23938958, 45.20131042, 42.28909167, 39.88266667, 45.41511667, 46.03584583, 46.03584583, 46.03584583, 44.99083125, 41.90855, 39.51920625, 39.59052708, 46.29081042, 46.29081042, 46.29081042, 45.23050417, 41.66845833, 39.28972083, 46.16619167, 46.78650417, 46.78650417, 46.78650417, 45.72059583, 41.61029375, 39.23586667, 46.6483125, 47.27002292, 47.27002292, 47.27002292, 46.18998125, 41.73563958, 39.35924792, 47.15570833, 47.77963542, 47.77963542, 47.77963542, 46.68295625, 42.04450833, 39.65987708, 47.68838542, 48.31532708, 48.31532708, 48.31532708, 47.19953333, 42.53688125, 40.13774375, 48.22966667, 48.86035208, 48.86035208, 48.86035208, 47.72362083, 43.0982875, 40.68222917, 48.59889375, 49.23307708, 49.23307708, 49.23307708, 48.080825, 43.48469792, 41.05761458, 48.7210625, 49.35807292, 49.35807292, 49.35807292, 48.19872292, 43.68674167, 41.25481667, 48.59616667, 49.235325, 49.235325, 49.235325, 48.07730833, 43.70441458, 41.27384375, 48.244225, 48.8848, 48.8848, 48.8848, 47.73600833, 43.57809375, 41.15377083, 47.88236042, 48.52267292, 48.52267292, 48.52267292, 47.38525, 43.39387083, 40.97789375, 47.60075208, 48.23875417, 48.23875417, 48.23875417, 47.11244792, 43.15504583, 40.749425, 47.39938958, 48.03302083, 48.03302083, 48.03302083, 46.91758958, 42.861625, 40.46835417, 47.28241042, 47.9100875, 47.9100875, 47.9100875, 46.80447708, 42.5552625, 40.17433958, 40.17399583, 47.85678125, 47.85678125, 47.85678125, 46.74513542, 42.3246875, 39.95186875, 47.4543125, 48.08268333, 48.08268333, 48.08268333, 46.96338958, 42.17330833, 39.80418125, 47.76176458, 48.39889375, 48.39889375, 48.39889375, 47.25246458, 42.10112292, 39.73129583, 48.21689167, 48.86836667, 48.86836667, 48.86836667, 47.68127708, 42.10815208, 39.73319375, 48.82369167, 49.494975, 49.494975, 49.494975, 48.25311667, 42.28407708, 39.89595208, 49.62545208, 50.32053333, 50.32053333, 50.32053333, 49.003575, 42.82007083, 40.40291875, 50.64016875, 51.36238125, 51.36238125, 51.36238125, 49.94746667, 43.72347083, 41.26117917, 51.86782708, 52.62054375, 52.62054375, 52.62054375, 51.08478542, 45.009575, 42.51391042, 52.694218, 54.07160625, 54.07160625, 54.07160625, 52.39361042, 46.42992708, 43.83725833, 54.64500208, 55.46198542, 55.46198542, 55.46198542, 53.63661042, 47.59869375, 44.946725, 55.84315, 56.686375, 56.686375, 56.686375, 54.71518958, 48.48396667, 45.78317292, 56.87996458, 57.74473333, 57.74473333, 57.74473333, 55.6293625, 49.08574375, 46.34662917, 57.75492708, 58.63667917, 58.63667917, 58.63667917, 56.38066042, 49.5323625, 46.76026458, 58.46225208, 59.35751875, 59.35751875, 59.35751875, 56.98600208, 50.09730208, 47.28654167, 58.9996, 59.90532917, 59.90532917, 59.90532917, 57.45240417, 50.79105833, 47.9356, 59.36694375, 60.28010833, 60.28010833, 60.28010833, 57.7798625, 51.61365417, 48.7074, 59.10752292, 60.01514792, 60.01514792, 60.01514792, 43.70312917, 42.98858958, 43.05765833, 43.19737917, 43.36140417, 43.50340208, 43.64655, 43.79083125, 53.64532917, 50.619275, 59.44871458, 60.365975, 60.365975, 60.365975, 57.92581458, 54.61548542, 51.53510833, 59.13325833, 60.04520208, 60.04520208, 60.04520208, 57.66232083, 54.96630417, 51.87153333, 58.64425625, 59.54701875, 59.54701875, 59.54701875, 57.21511042, 54.67821042, 51.6101875, 57.98170208, 58.87146875, 58.87146875, 58.87146875, 56.58422292, 53.75122083, 50.75102292, 57.16234792, 58.03564583, 58.03564583, 58.03564583, 55.78762083, 52.533275, 49.61746667, 56.36754375, 57.22510625, 57.22510625, 57.22510625, 55.0204125, 51.76600833, 48.89860625, 55.67260833, 56.51696458, 56.51696458, 56.51696458, 54.36364792, 51.47789792, 48.62097083, 55.07753958, 55.91119375, 55.91119375, 55.91119375, 53.81727708, 51.66896667, 48.78453125, 54.57965417, 55.40463958, 55.40463958, 55.40463958, 53.37890833, 52.01946042, 49.0931875, 54.14959375, 54.96304167, 54.96304167, 54.96304167, 53.02213333, 51.84792708, 48.91586667, 53.77517917, 54.57212292, 54.57212292, 54.57212292, 52.73600625, 51.12816042, 48.22829792, 53.45640417, 54.23190417, 54.23190417, 54.23190417, 52.52052083, 49.84336042, 47.29988913, 52.93666875, 53.68136875, 53.68136875, 53.68136875, 48.0439125, 48.04391875, 45.32239792, 45.32238958, 53.60258125, 53.60258125, 53.60258125, 52.20119375, 46.04206667, 43.44282917, 52.67117708, 53.36421458, 53.36421458, 53.36421458, 52.09403542, 44.62742708, 42.11370625, 52.23178958, 52.9078375, 52.9078375, 52.9078375, 51.73128125, 43.82972292, 41.36283333, 51.65584792, 52.32353542, 52.32353542, 52.32353542, 51.19846458, 43.64896667, 41.19016667, 41.19016458, 51.48672708, 51.48672708, 51.48672708, 50.36800833, 43.83646667, 41.36234792, 50.17636875, 50.84350833, 50.84350833, 50.84350833, 49.73565417, 43.86226667, 41.38198333, 49.36378125, 50.02800833, 50.02800833, 50.02800833, 48.93092917, 43.70599583, 41.22995833, 48.51150208, 49.1700125, 49.1700125, 49.1700125, 48.0895625, 43.36763958, 40.90627292, 40.906275, 48.15230833, 48.15230833, 48.15230833, 47.08946667, 42.96389375, 40.5222625, 46.9494625, 47.59293333, 47.59293333, 47.59293333, 46.55122917, 42.74343125, 40.31522917, 46.52954375, 47.16905417, 47.16905417, 47.16905417, 46.13615208, 42.71581667, 40.29430208, 46.37870833, 47.01715, 47.01715, 47.01715, 45.98468958, 42.88102917, 40.45947708, 46.49697083, 47.13723333, 47.13723333, 47.13723333, 46.09687917, 43.23910625, 40.81074375, 46.8639375, 47.50870833, 47.50870833, 47.50870833, 46.45270625, 43.71505833, 41.27474167, 47.25883542, 47.90856458, 47.90856458, 47.90856458, 46.8356625, 44.14911875, 41.69507708, 47.58996875, 48.24415833, 48.24415833, 48.24415833, 47.15579583, 44.53516042, 42.06575417, 47.8573375, 48.51547083, 48.51547083, 48.51547083, 47.41313125, 44.87315417, 42.38676875, 48.06164792, 48.723025, 48.723025, 48.723025, 47.6083875, 45.02209375, 42.5261875, 48.21071042, 48.87226667, 48.87226667, 48.87226667, 47.7496375, 44.68145208, 42.20286667, 48.30776458, 48.9654375, 48.9654375, 48.9654375, 47.84024792, 43.83966667, 41.40599167, 48.3528, 49.00253958, 49.00253958, 49.00253958, 47.88019792, 42.49674375, 40.13555208, 40.13555, 48.94366667, 48.94366667, 48.94366667, 47.8227375, 41.1157375, 38.82845417, 48.64567083, 49.27964167, 49.27964167, 49.27964167, 48.15894375, 40.6835375, 38.41570417, 49.29120208, 49.93097708, 49.93097708, 49.93097708, 48.78662083, 41.23808958, 38.93313333, 50.30837083, 50.96440208, 50.96440208, 50.96440208, 49.77787292, 42.77939167, 40.38075417, 51.69717292, 52.379925, 52.379925, 52.379925, 51.13269583, 45.30745833, 42.75855, 53.39522083, 54.11401042, 54.11401042, 54.11401042, 52.7885, 48.10732917, 45.39404375, 54.72666458, 55.47846042, 55.47846042, 55.47846042, 54.06721875, 49.65534792, 46.85412917, 55.410725, 56.187425, 56.187425, 56.187425, 54.68721667, 49.89293542, 47.08366875, 55.44742083, 56.24089583, 56.24089583, 56.24089583, 54.64846458, 48.83382917, 46.12703542, 54.40841, 55.72373542, 55.72373542, 55.72373542, 54.034625, 47.24798542, 44.61309792, 54.74033333, 55.55533958, 55.55533958, 55.55533958, 53.75193958, 46.9054875, 44.29879167, 55.2841375, 56.1175875, 56.1175875, 56.1175875, 54.17684583, 47.85907917, 45.20223125, 56.55211458, 57.41049792, 57.41049792, 57.41049792, 55.30934792, 50.1087625, 47.32339583, 58.463875, 59.35275417, 59.35275417, 59.35275417, 57.07254792, 52.97935625, 50.02681458, 60.14846667, 61.06349167, 61.06349167, 61.06349167, 58.6335625, 55.03196667, 51.95817917, 61.24409792, 62.17682292, 62.17682292, 62.17682292, 59.6464375, 56.21123958, 53.06540417, 61.75078125, 62.69273958, 62.69273958, 62.69273958, 60.1111625, 44.9026875, 44.15453542, 44.28999167, 44.44517292, 44.57418333, 44.69997917, 44.82256042, 44.97941458, 45.0375875, 45.14490208, 61.85667917, 61.85667917, 61.85667917, 59.31097917, 54.81995208, 51.73763958, 60.26981667, 61.19498333, 61.19498333, 61.19498333, 58.7129875, 53.7900875, 50.76854792, 59.54390833, 60.45763333, 60.45763333, 60.45763333, 58.03216042, 52.88568125, 49.92437083, 58.85806042, 59.75924375, 59.75924375, 59.75924375, 57.38955, 52.10674167, 49.20512292, 58.21822708, 59.10609375, 59.10609375, 59.10609375, 56.79146667, 51.58124167, 48.72606875, 57.68907292, 58.56618333, 58.56618333, 58.56618333, 56.30635208, 51.581975, 48.73290833, 57.29744167, 58.16774583]]).T

    T = 730
    sigma = 0.04 # daily volatility
    alpha = 0.06 # daily mean reversion
    K = forward_curve_arr[0,1] # Strike set to current price 
    
    mask = forward_curve_arr[:,0] <= T
    tau = forward_curve_arr[mask,0]
    forward_prices = forward_curve_arr[mask,1]
    
    # Helpers for analytical pricing of European options 
    w = 0.5 * sigma**2 * (1 - np.exp(-2*alpha*tau)) / alpha
    h = np.zeros(shape=w.shape)
    h[1:] = (np.log(forward_prices[1:]) - np.log(K) + 0.5 * w[1:]) / np.sqrt(w[1:])
    
    # (Undiscounted) Analytical European call and put option prices for each step
    european_call_px_analytical = forward_prices * norm.cdf(h) - K * norm.cdf(h - np.sqrt(w))
    european_put_px_analytical = K * norm.cdf(-h + np.sqrt(w)) - forward_prices * norm.cdf(-h)
    european_call_px_analytical_avg = european_call_px_analytical[2:].mean()
    european_put_px_analytical_avg = european_put_px_analytical[2:].mean()
    print('European analytical call price:', round(european_call_px_analytical_avg,4))
    print('European analytical put price:', round(european_put_px_analytical_avg,4))
    assert abs(round(european_call_px_analytical_avg,4) - 2.0413) < 1e8 
    assert abs(round(european_put_px_analytical_avg,4) - 5.1363) < 1e8 

    nb_simulations_list = [int(v) for v in [2e3, 4e3, 10e3, 15e3, 20e3, 50e3, 100e3]]
    dt_list = [1, 0.25, 0.0625, 0.03125]
    
    for i, dt in enumerate(dt_list):
        monte_carlo_avgs = []   
        monte_carlo_errors = []
    
        for nb_simulations in nb_simulations_list:
        
            segments_per_day = int(1 / dt)
            spot_px = clewlow_strickland_1_factor(forward_curve_arr, nb_simulations, segments_per_day, T, alpha, sigma, K=K)
        
            european_call_px_path_mc = np.maximum(0, spot_px - K).mean(axis=1)
            european_put_px_path_mc = np.maximum(0, K - spot_px).mean(axis=1)    
            
            call_px = european_call_px_path_mc[2:].mean()
            put_px = european_put_px_path_mc[2:].mean()
            monte_carlo_avgs.append((call_px,put_px))
            
            call_error = 100 * abs((call_px / european_call_px_analytical_avg) - 1)
            put_error = 100 * abs((put_px / european_put_px_analytical_avg) - 1)
            monte_carlo_errors.append((call_error, put_error))
            
        table = PrettyTable()
        print('')
        print(f'Table {i+1}: Δt={dt}, σ=0.04 and α=0.06: Average of option prices over maturities 2 to 730 days.')
        table.add_column("Sims", nb_simulations_list)
        table.add_column("Call Monte £/MWh", [round(c,4) for (c,p) in monte_carlo_avgs])
        table.add_column("Call Error (%)", [round(c,2) for (c,p) in monte_carlo_errors])
        table.add_column("Put Monte £/MWh", [round(p,4) for (c,p) in monte_carlo_avgs])
        table.add_column("Put Error (%)", [round(p,2) for (c,p) in monte_carlo_errors])
        print(table)
            
            
    








